/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
#ifndef DNS_RESOLVER_H
#define DNS_RESOLVER_H

#include <string>

#include <boost/net/dns.hpp>
#include <boost/shared_ptr.hpp>

#include "dns/hostaddress.h"

//-----------------------------------------------------------------------------
// DnsResolver
//-----------------------------------------------------------------------------

/**
 * @brief The class which provides Domain Name resolution to IP addresses.
 */
class DnsResolver
{
public:
    DnsResolver(
            const std::string &dns_address,
            const std::string &nameserver,
            const boost::net::dns::type_t address_resource_record_type,
            const int resolved_ip_ttl_threshold
    );
    ~DnsResolver();

    bool resolve();

    int get_resolved_ip_count() const;
    std::string get_next_ip();
    bool have_up_to_date_ip() const;

private:
    bool handle_ip_address();

    boost::net::dns::rr_list_t get_answers_list(
            const std::string &host_dns_address,
            const std::string &name_server,
            const boost::net::dns::type_t message_type
    ) const;
    void fill_resolved_ip_list(
            const boost::net::dns::rr_list_t& answers_list,
            std::list<HostAddress> *resolved_host_address_list
    ) const;
    void append_resolved_ipv4(
            const shared_ptr<boost::net::dns::resource_base_t> resource_record,
            std::list<HostAddress> *resolved_host_address_list
    ) const;
    void append_resolved_ipv6(
            const shared_ptr<boost::net::dns::resource_base_t> resource_record,
            std::list<HostAddress> *resolved_host_address_list
    ) const;

private:
    /// The list of IPs available to the host DNS.
    HostAddressList ResolvedHostAddressList;
    /// The DNS of the host.
    const std::string HostDnsAddress;
    /// The address of the server which can resolve the host address.
    const std::string NameServer;
    /// 'A' resource records for the IPv4 addresses, or 'AAAA' resource records for the IPv6 addresses
    boost::net::dns::type_t AddressResourceRecord;
    /// Minimum time that IP has to be valid before creating new dns request
    const uint32_t ResolvedIpTtlThreshold;

};

//-----------------------------------------------------------------------------
// DnsResolverItem
//-----------------------------------------------------------------------------

typedef boost::shared_ptr<DnsResolver> DnsResolverItem;

#endif // DNS_RESOLVER_H
