/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#include "dns_neww/dnsmaster.h"

#include <logfunc.hpp>
#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/asio/placeholders.hpp>

using boost::bind;
using boost::posix_time::seconds;
using I2n::Logger::GlobalLogger;

DnsMasterItem DnsMaster::TheOnlyInstance;

void DnsMaster::create_master(const IoServiceImte &io_serv,
                              const boost::ip::address &name_server,
                              const std::string &cache_file)
{
    if (TheOnlyInstance)
    {
        GlobalLogger.warning()
            << "Attempting to create another DnsMaster instance!";
        return false;
    }

    TheOnlyInstance.reset( new DnsMaster(io_serv, name_server, cache_file) );
    return true;
}


DnsMaster::DnsMaster(const IoServiceItem &io_serv,
                     const boost::ip::address &name_server,
                     const std::string &cache_file)
    : IoService( io_serv )
    , NameServer( name_server )
    , SaveTimer( *io_serv )
    , CacheFile( cache_file )
    , ResolverMap()
    , DataCache()
{
    // load cache from file
    load_from_cachefile();

    // schedule next save
    (void) SaveTimer.expires_from_now( seconds( SaveTimerSeconds ) );
    SaveTimer.async_wait( bind( &DnsMaster::schedule_save, this,
                                boost::asio::placeholders::error ) );
}


DnsMasterItem& DnsMaster::get_instance()
{
    if ( !TheOnlyInstance )
        GlobalLogger.error()
            << "Request to return DnsMaster instance before creating it!";
    return TheOnlyInstance;
}



DnsMaster::~DnsMaster()
{
    // save one last time without re-scheduling the next save
    save_to_cachefile();

    // cancel save timer
    SaveTimer.cancel();
}

void DnsMaster::schedule_save(const boost::system::error_code &error)
{
    // just in case: ensure SaveTimer is cancelled
    SaveTimer.cancel();  // (will do nothing if already expired/cancelled)

    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
    {
        GlobalLogger.error() << "SaveTimer was cancelled "
                             << "--> no save and no re-schedule of saving!";
        return;
    }
    else if (error)
    {
        GlobalLogger.error() << "Received error " << error
                             << " in schedule_save "
                             << "--> no save now but re-schedule saving";
    }
    else
        save_to_cachefile();

    // schedule next save
    (void) SaveTimer.expires_from_now( seconds( SaveTimerSeconds ) );
    SaveTimer.async_wait( bind( &DnsMaster::schedule_save, this,
                                boost::asio::placeholders::error ) );
}

void DnsMaster::save_to_cachefile() const
{   // now do the saving
    // TODO (see trusted_net_helper, boost serialization, xml)
    GlobalLogger.error() << "Actual saving not implemented yet!";
}


bool DnsMaster::load_from_cachefile()
{
    // TODO: some boost serialization and xml stuff, see trusted_net_helper
    GlobalLogger.error() << "Actual loading not implemented yet!";
    return true;
}

ResolverItem DnsMaster::get_resolver_for( const std::string &hostname )
{
    DnsMasterItem master = get_instance();
    if ( master->ResolverMap.count(hostname) == 0 )
    {   // need to create a resolver

        // check if it is an ip address, so can create a simple pseudo resolver
        if ( master->is_ip(hostname) )
        {
            ResolverItem new_resolver( new IpPseudoResolver(hostname) );
            master->ResolverMap[hostname] = new_resolver;
        }
        else
        {
            ResolverItem new_resolver( new DnsResolver(IoService,
                                                       NameServer,
                                                       hostname) );
            master->ResolverMap[hostname] = new_resolver;
        }
    }
    return master->ResolverMap[hostname];
}

/**
 * return true if given hostname string actually is an IP
 *
 * delegates decision to boost::ip::address::from_string
 */
bool DnsMaster::is_ip(const std::string &hostname) const
{
    try
    {
        boost::ip::address ip = boost::ip::address::from_string(hostname);
        return ip.is_v4() || ip.is_v6();
    }
    catch ( const std::exception &ex )
    {
        return false;
    }
}

// (created using vim -- the world's best text editor)

