/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#include "dns_neww/dnsresolver.h"

#include <logfunc.hpp>
using I2n::Logger::GlobalLogger;

namespace Config
{
    const int ResolveTimeoutSeconds = 10;
    const int DNS_PORT = 53;
    const int UniqueID = 0xaffe;
}

DnsResolver::DnsResolver(IoServiceItem &io_serv,
                         const boost::ip::address &name_server,
                         const std::string &hostname)
    : Hostname( hostname )
    , Socket( *io_serv, ip::udp::endpoint(ip::udp::v4(), 0) )
    , ReplyBuffer()
    , NameServer( name_server, DNS_PORT )
    , ResolveTimeoutTimer( *io_serv )
    , PauseBeforeRetryTimer( *io_serv )
    , StaleDataLongtermTimer( *io_serv )
    , CallbackList()
    , host_ttl_time_list::const_iterator()
    , RetryCount( 0 )
    , IsResolving( false )
{
    // add name server to resolver
    resolver.addServer( nameServer );
}

/**
 * copied here code from boost::net::dns::resolve.hpp, since want async 
 * operation and that is used only internally, there
 * --> give credit to Andreas Haberstroh (andreas at ibusy dot com)
 *     from https://github.com/softwareace/Boost.DNS
 */
DnsResolver::async_resolve(const callback_type callback, const int cname_count)
{
    // remember callback
    CallbackList.push_back(callback, cname_count);

    // check if resolving already
    if (isResolving)
    {
        GlobalLogger().info() << "Call to async_resolve(" << Hostname
                              << ") ignored since resolving already";
        return;
    }

    // just to be sure: cancel timers
    ResolveTimeoutTimer.cancel();
    PauseBeforeRetryTimer.cancel();
    StaleDataLongtermTimer.cancel();

    // create DNS request
    boost::net::dns::message dns_message( host_dns_address,
                                          boost::net::dns::type_all );
    dns_message.recursive(false);
    dns_message.action(dns::message::query);
    dns_message.opcode(dns::message::squery);
    dns_message.id(UniqueID);
    dns_buffer_t request_buffer;
    dns_message.encode(request_buffer);

    // setup receipt of reply
    Socket.async_receive_from(
            boost::asio::buffer(ReceiveBuffer.get_array()),
            NameServer,
            boost::bind( &DnsResolver::handle_dns_result, this,
                         boost::asio::placeholders::error,
                         boost::asio::placeholders::bytes_transferred)
    )

    // schedule timeout
    (void) ResolveTimeoutTimer.expires_from_now(seconds(ResolveTimeoutSeconds));
    ResolveTimeout.async_wait( bind( &DnsResolver::wait_timer_timeout_handler,
                                     this, boost::asio::placeholders::error) );

    // send dns request
    Socket.send_to( boost::asio::buffer(request_buffer.get_array()),
                    NameServer );
}


void handle_dns_result(const boost::system::error_code &error,
                       const std::size_t bytes_transferred)
{
    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
    {
        GlobalLogger.info() << "DNS resolve operation was cancelled";
        bool was_success = false;
        bool inc_cname = false;
        finalize_resolve(was_success, inc_cname);
    }
    else if (error)
    {
        GlobalLogger.info() << "DNS resolve resulted in error " << error
                            << " --> treat like unavailable";
        handle_unavailable();
        return;
    }

    // next 3(+1) lines copied from boost/net/dns/resolver.hpp:
    // clamp the recvBuffer with the number of bytes transferred or decode buffr
    ReceiveBuffer.length(bytes_transferred);
    message result_message();
    result_message.decode( ReceiveBuffer );

    if (result_message.answers().size == 0)
        handle_unavailable();

    // loop over answers, remembering ips and cnames
    typedef boost::shared_ptr<resource_base_t> rr_item_type;
    typedef boost::shared_ptr<cname_resource> cname_item_type;
    std::vector<HostAddress> ip_list;
    std::vector<cname_item_type> cname_list;
    BOOST_FOREACH( rr_item_type rr_item, result_message.answers() )
    {
        GlobalLogger.debug() << std::showbase << std::hex << rr_item->rtype()
                             << ": ";
        uint32_t ttl = rr_item->ttl();
        type_t rr_type = rr_item->rtype();

        if (rr_type == boost::net::dns::type_a)
        {    // 'A' resource records carry IPv4 addresses
            std::string ip = ( dynamic_cast<a_resource *> (rr_item.get()) )
                             ->address();
            ip_list.push_back( HostAddress(ip, ttl) );
        }
        else if (rr_type == boost::net::dns::type_a6)
        {   // 'AAAA' resource records carry IPv6 addresses
            std::string ip = ( dynamic_cast<a6_resource *> (rr_item.get()) )
                             ->address();
            ip_list.push_back( HostAddress(ip, ttl) );
        }
        else if (rr_type == boost::net::dns::type_cname)
        {   // 'CNAME' resource records that carry aliases
            cname_list.push_back(dynamic_cast<cname_item_type>(rr_item));
        }
        else if (rr_type == boost::net::dns::type_ns)
            GlobalLogger.debug() << "NS resource";
        else if (rr_type == boost::net::dns::type_soa)
            GlobalLogger.debug() << "SOA resource";
        else if (rr_type == boost::net::dns::type_ptr)
            GlobalLogger.debug() << "ptr resource";
        else if (rr_type == boost::net::dns::type_hinfo)
            GlobalLogger.debug() << "hinfo resource";
        else if (rr_type == boost::net::dns::type_mx)
            GlobalLogger.debug() << "mx resource";
        else if (rr_type == boost::net::dns::type_txt)
            GlobalLogger.debug() << "txt resource";
        else if (rr_type == boost::net::dns::type_srv)
            GlobalLogger.debug() << "srv resource";
        else if (rr_type == boost::net::dns::type_axfr)
            GlobalLogger.debug() << "axfr resource";
        else
            GlobalLogger.debug() << "unknown resource type";
    }

    GlobalLogger.info() << "Have " << ip_list.size() << " IPs and "
                        << cname_list.size() << " CNAMEs";

    // We expect either one single CNAME and no IPs or a list of IPs.
    // But deal with other cases as well
    if (ip_list.empty() && cname_list.empty())
        handle_unavailable();   // we just got crap, this is a dead end
    else if ( !ip_list.empty() && !cname_list.empty())
        GlobalLogger.warning() << "Have CNAMEs AND IPs --> deal with both!";

    BOOST_FOREACH( const cname_resource &cname, cname_list )
        handle_cname(cname);  // will schedule another DNS call

    if ( !ip_list.empty() )
        handle_ips(ip_list);
}


void DnsResolver::handle_ips(const boost::system::error_code &error,
                             const IpTtlVec &ips)
{
    if (error)
        GlobalLogger.warning() << "Received error " << error
                               << " in handle_ips --> exit!";
    // save in cache
    update_master(ips);

    // clean up
    bool was_success = true;
    bool inc_cname = false;
    finalize_resolve(was_success, inc_cname);
}


void DnsResolver::handle_unavailable()
{
    // schedule new attempt in quite a while
    StaleDataLongtermTimer.expires_from_now(seconds(StaleDataLongtermSeconds));
    StaleDataLongtermTimer.async_wait(
            bind( &DnsResolver::wait_timer_timeout_handler,
                  this, boost::asio::placeholders::error
            )
    );

    // for now, admit failure
    bool was_success = false;
    bool inc_cname = false;
    finalize_resolve(was_success, inc_cname);
}

void DnsResolver::handle_cname(const std::string &canonical_name)
{
    // get resolver for canonical name
    ResolverItem resolver = DnsMaster.get_instance()
                            ->get_resolver_for(canonical_name);
    resolver->async_resolve(
            bind( &DnsResolver::cname_resolve_callback,
                  this, boost::asio::placeholders::error, _1, _2
            )
    );

    stop_trying();
}


void cname_resolve_callback(const boost::system::error_code &error,
                            const bool was_success,
                            const int cname_count)
{
    if (error)
        GlobalLogger.error() << "todo"; // TODO
    if (!was_success)
        GlobalLogger.error() << "todo"; // TODO

    // get all results
    host_ttl_time_list results = DnsMaster.get_instance()->get_cached_results();
}


DnsResolver::get_next_ip()
{
    // get cached data
    host_ttl_time_list cached_data = DnsMaster::get_instance()
                                     ->get_cached_results(Hostname);


}

// (created using vim -- the world's best text editor)

