/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "host/pingrotate.h"

#include <boost/bind.hpp>
#include <boost/foreach.hpp>

#include "boost_assert_handler.h"
#include "dns/dnsresolverfactory.h"
#include "host/pingerfactory.h"

using namespace std;
using boost::function;
using boost::shared_ptr;

//-----------------------------------------------------------------------------
// PingRotate
//-----------------------------------------------------------------------------

/**
 * @brief Parameterized constructor.
 *
 * @param io_serv The one @c io_service object that controls async processing
 * @param network_interface The name of the network interface from where to
 * dispatch the pings.
 * @param destination_address The remote address to ping.
 * @param destination_port The remote port to ping.
 * @param nameserver Server to resolve the addresses.
 * @param protocol_list A list of protocols to be used to ping the
 * host. The protocols will be used in the order they are in the list.
 */
PingRotate::PingRotate(
        const IoServiceItem io_serv,
        const string &network_interface,
        const string &destination_address,
        const uint16_t destination_port,
        const string &nameserver,
        const int resolved_ip_ttl_threshold,
        const int ping_reply_timeout,
        const PingProtocolList &protocol_list
) :
    IoService( io_serv ),
    NetworkInterfaceName( network_interface ),
    IpList(),
    DestinationAddess( destination_address ),
    DestinationPort( destination_port ),
    Nameserver( nameserver ),
    ResolvedIpTtlThreshold( resolved_ip_ttl_threshold ),
    PingReplyTimeout( ping_reply_timeout ),
    ProtocolRotate( protocol_list.size() ),
    Ping(),
    PingDoneCallback()
{
    BOOST_ASSERT( !network_interface.empty() );
    BOOST_ASSERT( !destination_address.empty() );
    BOOST_ASSERT( 0 < destination_port );
    BOOST_ASSERT( !nameserver.empty() );
    BOOST_ASSERT( 0 < protocol_list.size() );

    // fill circular buffer with protocols
    BOOST_FOREACH( const PingProtocol &prot, protocol_list )
        ProtocolRotate.push_back(prot);

    init_ping_protocol();
}

/**
 * @brief Destructor.
 */
PingRotate::~PingRotate()
{
}

/**
 * @brief Ping a destination address from an available local source.
 *
 * @param done_handler Done handler will be called on successful ping or timeout.
 *
 * @return void.
 */
void PingRotate::ping( function<void(bool)> ping_done_callback )
{
    BOOST_ASSERT( ( 0 < DestinationPort ) && ( DestinationPort < numeric_limits<uint16_t>::max() ) );

    set_ping_done_callback( ping_done_callback );

    update_ping_protocol();

    string destination_ip = IpList->get_next_ip();

    Ping->ping(
            destination_ip,
            DestinationPort,
            boost::bind(&PingRotate::ping_done_handler, this, _1)
    );
}

void PingRotate::stop_pinging()
{
    Ping->stop_pinging();
}

bool PingRotate::resolve_ping_address()                                                         //lint !e1762
{
    return IpList->resolve();
}

int PingRotate::get_resolved_ip_count() const
{
    return IpList->get_resolved_ip_count();
}

bool PingRotate::have_up_to_date_ip() const
{
    return IpList->have_up_to_date_ip();
}

void PingRotate::set_ping_done_callback( function<void(bool)> ping_done_callback )
{
    PingDoneCallback = ping_done_callback;
}

void PingRotate::ping_done_handler( bool ping_success ) const
{
    PingDoneCallback( ping_success );
}

void PingRotate::init_ping_protocol()
{
    get_next_ping_protocol();
}

void PingRotate::update_ping_protocol()
{
    if ( can_change_ping_protocol() )
    {
        get_next_ping_protocol();
    }
}

void PingRotate::get_next_ping_protocol()
{
    PingProtocol ping_protocol = ProtocolRotate.front();
    ProtocolRotate.pop_front();
    ProtocolRotate.push_back(ping_protocol);

    Ping = PingerFactory::createPinger( ping_protocol, IoService, NetworkInterfaceName, PingReplyTimeout );

    update_dns_resolver( ping_protocol );
}

bool PingRotate::can_change_ping_protocol() const
{
    // TODO can_change_ping_protocol() and get_next_ping_protocol() may be implemented in a Algorithm
    // class that can be exchanged in this class to provide an algorithm neutral class
    return true;
}

void PingRotate::update_dns_resolver( PingProtocol current_protocol )
{
    IpList = DnsResolverFactory::createResolver( DestinationAddess, Nameserver, ResolvedIpTtlThreshold, current_protocol );

    // when the protocol change, there is a chance that the IP version required by the new
    // protocol differ from the previous one, thus it is required to resolve the address again.
    // FIXME add some intelligence here to decide if the new protocol requires to update the DNS
    resolve_ping_address();                                                                         //lint !e534
}
